﻿using AIWolf.Lib;
using System;
using System.Collections.Generic;
using System.Linq;
using WhiteWolf.Arrangement;
using WhiteWolf.Guess;

namespace WhiteWolf.ActionPlan
{
    public class SampleActionPlanner : IActionPlanner
    {
        public IActionPlan GetPlan(ITotalityGameInfo gameInfo, ITotalGuess guesses)
        {
            var partRequests = GetPartRequests(gameInfo, guesses);
            var groups = partRequests.GroupBy(request => request.Agent).ToList();

            Dictionary<Agent, double> voteScores = new Dictionary<Agent, double>();
            Dictionary<Agent, double> attackScores = new Dictionary<Agent, double>();
            Dictionary<Agent, double> guardScores = new Dictionary<Agent, double>();
            Dictionary<Agent, double> divineScores = new Dictionary<Agent, double>();
            foreach (Agent agent in gameInfo.AgentList)
            {
                voteScores[agent] = groups.FirstOrDefault(group => group.Key.Equals(agent))?.Select(request => request.Vote).Aggregate((r1, r2) => r1 * r2) ?? 1.0;
                attackScores[agent] = groups.FirstOrDefault(group => group.Key.Equals(agent))?.Select(request => request.Attack).Aggregate((r1, r2) => r1 * r2) ?? 1.0;
                guardScores[agent] = groups.FirstOrDefault(group => group.Key.Equals(agent))?.Select(request => request.Guard).Aggregate((r1, r2) => r1 * r2) ?? 1.0;
                divineScores[agent] = groups.FirstOrDefault(group => group.Key.Equals(agent))?.Select(request => request.Divine).Aggregate((r1, r2) => r1 * r2) ?? 1.0;
            }

            var voteTarget = voteScores.OrderByDescending(kv => kv.Value).First().Key;
            var attackTarget = attackScores.OrderByDescending(kv => kv.Value).First().Key;
            var guardTarget = guardScores.OrderByDescending(kv => kv.Value).First().Key;
            var divineTarget = divineScores.OrderByDescending(kv => kv.Value).First().Key;

            return new ActionPlan(voteTarget, attackTarget, guardTarget, divineTarget, partRequests);
        }


        private List<IPartActionRequest> GetPartRequests(ITotalityGameInfo gameInfo, ITotalGuess guesses)
        {
            var partRequests = new List<IPartActionRequest>();

            // 自分が死んでいたら何もできないので空を返す
            if (!gameInfo.AliveAgentList.Contains(gameInfo.Me))
            {
                return partRequests;
            }


            // 自分は対象にできない（占い対象にはできるけどしない）
            partRequests.Add(new PartActionRequest(gameInfo.Me, 0.0, 0.0001, 0.0, 0.0));

            // 死人は対象にできない
            gameInfo.DayGameInfo.Values.Where(day => day.AttackDeadAgent != null).SelectMany(day => day.AttackDeadAgent).ToList()
                .ForEach(agent => partRequests.Add(new PartActionRequest(agent, 0.0, 0.0, 0.0, 0.0)));
            gameInfo.DayGameInfo.Values.Where(day => day.ExecuteAgent != null).Select(day => day.ExecuteAgent).ToList()
                .ForEach(agent => partRequests.Add(new PartActionRequest(agent, 0.0, 0.0, 0.0, 0.0)));

            // 仲間狼は襲撃できない、投票あまりしない
            if (gameInfo.GameSetting.PlayerNum == 15 && gameInfo.MyRole == Role.WEREWOLF)
            {
                partRequests.AddRange(gameInfo.WerewolfAgentList.Select(wolf => new PartActionRequest(wolf, 0.7, 1.0, 1.0, 0.0)));
            }

            // 村側の行動原則に従った行動（人外は５人以上でのみ行うが、投票度は軽視する）
            if (gameInfo.MyRole.GetTeam() == Team.VILLAGER)
            {
                partRequests.AddRange(GetVillagerPartRequests(gameInfo, guesses)
                    .Select(req => new PartActionRequest(req.Agent, Math.Pow(req.Vote, 0.5), req.Divine, req.Guard, req.Attack)));
            }
            else if (gameInfo.AliveAgentList.Count > 4)
            {
                partRequests.AddRange(GetVillagerPartRequests(gameInfo, guesses));
            }

            // 人狼の行動
            if (gameInfo.MyRole == Role.WEREWOLF)
            {
                partRequests.AddRange(GetWerewolfPartRequests(gameInfo, guesses));
            }


            // 票合わせ、吊り逃れの行動
            if (gameInfo.MyRole.GetTeam() == Team.VILLAGER)
            {
                partRequests.AddRange(GetVoteStackPartRequests(gameInfo, guesses));
            }
            if (gameInfo.MyRole == Role.WEREWOLF)
            {
                partRequests.AddRange(GetVoteStackPartRequests(gameInfo, guesses).Select(req => new PartActionRequest(req.Agent, Math.Pow(req.Vote, 1.5), 1.0, 1.0, 1.0)));
            }
            if (gameInfo.MyRole == Role.POSSESSED)
            {
                partRequests.AddRange(GetVoteStackPartRequests(gameInfo, guesses).Select(req => new PartActionRequest(req.Agent, Math.Pow(req.Vote, 0.5), 1.0, 1.0, 1.0)));
            }

            // 狂視点の推理を反映
            if (gameInfo.MyRole == Role.POSSESSED)
            {
                var posGuess = new TotalGuess(gameInfo.AgentList, gameInfo.SelfRealRoleViewTrustInfo.Patterns.Select(p => (p, 1.0)));

                // 村の可能性が無い人には投票しない
                partRequests.AddRange(posGuess.likelyVillagerPattern.Where(a => a.Value.score == 0.0).Select(p => new PartActionRequest(p.Key, 0.0001, 1.0, 1.0, 1.0)));
                // 狼の可能性が無い人には投票する
                partRequests.AddRange(posGuess.LikelyWolfPattern.Where(a => a.Value.score == 0.0).Select(p => new PartActionRequest(p.Key, 2.0, 1.0, 1.0, 1.0)));
            }



            // 勝率高いの殺しておく

            // PP

            // ライン切り


            // 再投票



            return partRequests;
        }


        private List<IPartActionRequest> GetVillagerPartRequests(ITotalityGameInfo gameInfo, ITotalGuess guesses)
        {
            var partRequests = new List<IPartActionRequest>();


            // 怪しさに応じて対象にする
            foreach (Agent agent in gameInfo.AgentList)
            {
                var (wPattern, wScore) = guesses.LikelyWolfPattern[agent];
                var (pPattern, pScore) = guesses.LikelyPossessedPattern[agent];
                var (vPattern, vScore) = guesses.LikelyVillagerPattern[agent];

                var wolfsideScore = wScore + pScore / 3;

                // TODO ラス縄で狂人吊りそうなので調整する
                double voteScore = (wolfsideScore + 0.01) / (vScore + 0.01);
                voteScore = Math.Pow(Math.Min(Math.Max(voteScore, 0.1), 5.0), 0.8);
                double attackScore = Math.Pow(1.0 / voteScore, 0.5);
                double divineScore = Math.Pow(voteScore, 0.5);
                double guardScore = Math.Pow(1.0 / voteScore, 0.5);

                if (vScore == 0.0)
                {
                    divineScore = 0.2;
                }
                if (wScore == 0.0)
                {
                    divineScore = 0.01;
                }

                partRequests.Add(new PartActionRequest(agent, voteScore, divineScore, guardScore, attackScore));
            }

            // 15人村4日目になるまで
            if (gameInfo.GameSetting.PlayerNum == 15 && gameInfo.TodayGameInfo.Day < 4)
            {
                // 占い師に投票しない
                foreach (var co in gameInfo.ComingOuts.Where(co => co.IsEnable && co.Role == Role.SEER))
                {
                    double voteScore = 0.1;
                    double attackScore = 3.0;
                    double divineScore = 0.1;
                    double guardScore = 2.0;
                    partRequests.Add(new PartActionRequest(co.Agent, voteScore, divineScore, guardScore, attackScore));
                }
                // 白に投票しない
                foreach (var judge in gameInfo.SeerJudges.Where(judge => judge.IsEnable && judge.Result == Species.HUMAN))
                {
                    double voteScore = 0.3;
                    double attackScore = 2.0;
                    double divineScore = 0.7;
                    double guardScore = 1.4;
                    partRequests.Add(new PartActionRequest(judge.Target, voteScore, divineScore, guardScore, attackScore));
                }
                // 黒に投票する
                foreach (var agent in gameInfo.AliveAgentList.Where(agent => gameInfo.SeerJudges.Any(judge => judge.IsEnable && judge.Target.Equals(agent) && judge.Result == Species.WEREWOLF)))
                {
                    double voteScore = 2.0;
                    double attackScore = 0.1;
                    double divineScore = 0.3;
                    double guardScore = 0.1;
                    partRequests.Add(new PartActionRequest(agent, voteScore, divineScore, guardScore, attackScore));
                }
            }

            // 5人村初日のみ
            if (gameInfo.GameSetting.PlayerNum == 5 && gameInfo.TodayGameInfo.Day < 2)
            {
                // 初日のみ占い師に投票しない
                foreach (var co in gameInfo.ComingOuts.Where(co => co.IsEnable && co.Role == Role.SEER))
                {
                    double voteScore = 0.5;
                    double attackScore = 2.0;
                    double divineScore = 0.2;
                    double guardScore = 1.0;
                    partRequests.Add(new PartActionRequest(co.Agent, voteScore, divineScore, guardScore, attackScore));
                }
                // 初日のみ白に投票しない
                foreach (var judge in gameInfo.SeerJudges.Where(judge => judge.IsEnable && judge.Result == Species.HUMAN))
                {
                    double voteScore = 0.8;
                    double attackScore = 1.5;
                    double divineScore = 0.7;
                    double guardScore = 1.0;
                    partRequests.Add(new PartActionRequest(judge.Target, voteScore, divineScore, guardScore, attackScore));
                }
                // 初日のみ黒に投票する
                foreach (var agent in gameInfo.AliveAgentList.Where(agent => gameInfo.SeerJudges.Any(judge => judge.IsEnable && judge.Target.Equals(agent) && judge.Result == Species.WEREWOLF)))
                {
                    double voteScore = 1.5;
                    double attackScore = 0.3;
                    double divineScore = 0.3;
                    double guardScore = 1.0;
                    partRequests.Add(new PartActionRequest(agent, voteScore, divineScore, guardScore, attackScore));
                }
            }

            return partRequests;
        }

        private List<IPartActionRequest> GetWerewolfPartRequests(ITotalityGameInfo gameInfo, ITotalGuess guesses)
        {
            var partRequests = new List<IPartActionRequest>();

            // 5人村 占い師2人いれば残す
            if (gameInfo.GameSetting.PlayerNum == 5)
            {
                var seers = gameInfo.ComingOuts.Where(co => co.IsEnable && co.Role == Role.SEER).Select(co => co.Agent).ToList();
                if (seers.Count == 2 && seers.All(seer => gameInfo.AliveAgentList.Contains(seer)))
                {
                    partRequests.AddRange(seers.Select(seer => new PartActionRequest(seer, 1.0, 1.0, 1.0, 0.01)));
                }
            }

            // 前日が平和なら同じ場所は襲撃しない
            if (gameInfo.YesterdayGameInfo?.TryAttackAgent != null && !gameInfo.YesterdayGameInfo.AttackDeadAgent.Any())
            {
                partRequests.Add(new PartActionRequest(gameInfo.YesterdayGameInfo.TryAttackAgent, 1.0, 1.0, 1.0, 0.2));
            }

            // 人狼に投票した人を襲撃する
            if (gameInfo.GameSetting.PlayerNum == 15 && gameInfo.TodayGameInfo.VoteList.Any())
            {
                var saidVote = gameInfo.TodayGameInfo.VoteList.First();
                partRequests.AddRange(saidVote.VoteMap.Where(vote => gameInfo.WerewolfAgentList.Contains(vote.Value)).Select(vote => new PartActionRequest(vote.Key, 1.0, 1.0, 1.0, 1.2)));
            }

            return partRequests;
        }

        private List<IPartActionRequest> GetVoteStackPartRequests(ITotalityGameInfo gameInfo, ITotalGuess guesses)
        {
            var partRequests = new List<IPartActionRequest>();

            // 票合わせ
            var saidVote = new VoteCollection(gameInfo.AliveAgentList, gameInfo.TodayGameInfo.TalkList);
            var receiveVoteCountMap = gameInfo.AliveAgentList.ToDictionary(agent => agent, agent => saidVote.VoteMap.Count(vote => !vote.Key.Equals(gameInfo.Me) && agent.Equals(vote.Value)));
            var meReceiveVoteCount = receiveVoteCountMap[gameInfo.Me];
            var totalVoteCount = saidVote.VoteMap.Count(vote => vote.Value != null);
            if (totalVoteCount > 0)
            {
                foreach (var vc in receiveVoteCountMap.Where(vc => vc.Value != 0))
                {
                    double voteScore = 1.0 + 0.8 * vc.Value / totalVoteCount;
                    double attackScore = 0.9;
                    double divineScore = 0.98;
                    double guardScore = 0.8;
                    partRequests.Add(new PartActionRequest(vc.Key, voteScore, divineScore, guardScore, attackScore));
                }
            }

            // 吊り逃れ
            var maxReceiveVoteCount = receiveVoteCountMap.Max(count => count.Value);
            // 昨年のスコアに合わせる？
            if (meReceiveVoteCount == maxReceiveVoteCount & maxReceiveVoteCount >= 2)
            {
                var aaa = receiveVoteCountMap.Where(count => count.Value >= maxReceiveVoteCount - 1);
                partRequests.AddRange(aaa.Select(count => new PartActionRequest(count.Key, 1.5 + (maxReceiveVoteCount - count.Value) * 0.5, 1, 1, 1)));
            }
            if (meReceiveVoteCount == maxReceiveVoteCount - 1 & maxReceiveVoteCount >= 2)
            {
                var aaa = receiveVoteCountMap.Where(count => count.Value >= maxReceiveVoteCount - 1);
                partRequests.AddRange(aaa.Select(count => new PartActionRequest(count.Key, 1 + (maxReceiveVoteCount - count.Value) * 0.5, 1, 1, 1)));
            }

            // 間違った判定を出した占候補は噛まない
            partRequests.AddRange(gameInfo.SeerJudges
                .Where(judge => (judge.Result == Species.WEREWOLF) != gameInfo.WerewolfAgentList.Contains(judge.Target))
                .Select(judge => judge.Agent)
                .Distinct()
                .Select(seer => new PartActionRequest(seer, 0.8, 1.0, 1.5, 0.5)));


            return partRequests;
        }



    }
}
