﻿using AIWolf.Lib;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;

namespace WhiteWolf.Arrangement
{
    public class TotalityGameInfo : ITotalityGameInfo
    {

        public IGameInfo latestGameInfo { get; private set; }


        public IGameSetting GameSetting { get; }

        // TODO readonlyにする
        public SortedDictionary<int, IDayGameInfo> DayGameInfo { get; } = new SortedDictionary<int, IDayGameInfo>();

        public IDayGameInfo TodayGameInfo { get; private set; }

        public IDayGameInfo YesterdayGameInfo { get; private set; }

        public Agent Me => latestGameInfo.Agent;

        public Role MyRole => latestGameInfo.Role;

        public IReadOnlyList<Agent> AgentList => new ReadOnlyCollection<Agent>(latestGameInfo.AgentList);

        public IReadOnlyList<Agent> AliveAgentList => new ReadOnlyCollection<Agent>(latestGameInfo.AliveAgentList);

        public IViewpoint AllViewSystemInfo { get; private set; }

        public IViewpoint AllViewTrustInfo { get; private set; }

        public IViewpoint SelfViewSystemInfo { get; private set; }

        public IViewpoint SelfViewTrustInfo { get; private set; }

        public IViewpoint SelfRealRoleSystemtInfo { get; private set; }

        public IViewpoint SelfRealRoleViewTrustInfo { get; private set; }

        public IEnumerable<ComingOut> ComingOuts { get; private set; } = new List<ComingOut>();

        public IEnumerable<Judge> SeerJudges { get; private set; } = new List<Judge>();

        public IEnumerable<Judge> MediumJudges { get; private set; } = new List<Judge>();


        #region "内部ロジック"

        /// <summary>初期視点を生成するFactory</summary>
        Func<IGameSetting, IViewpoint> ViewPointFactoryMethod { get; set; } = ViewpointFactory.GetViewPoint;

        public IReadOnlyList<Agent> WerewolfAgentList => latestGameInfo.RoleMap.Where(map => map.Value == Role.WEREWOLF).Select(map => map.Key).ToList();


        #endregion


        public void Update(IGameInfo gameInfo)
        {
            // 最新のゲーム情報を保管
            latestGameInfo = gameInfo;

            // 日付が変わっているか
            var isDayUpdate = !DayGameInfo.ContainsKey(gameInfo.Day);

            if (isDayUpdate)
            {
                DayStart(gameInfo);
            }
            else
            {
                TodayGameInfo.Update(gameInfo);
            }

            // COと判定結果からの更新
            UpdateComingOutAndJudge();
        }

        /// <summary>
        /// 日付変更時(昼の開始時)の処理
        /// </summary>
        private void DayStart(IGameInfo gameInfo)
        {
            // 昨日の情報を更新
            YesterdayGameInfo = TodayGameInfo;
            if (YesterdayGameInfo != null && YesterdayGameInfo.Day + 1 == gameInfo.Day)
            {
                YesterdayGameInfo.Update(gameInfo);
            }

            // 日の情報を作成
            TodayGameInfo = new DayGameInfo(gameInfo);
            TodayGameInfo.Update(gameInfo);
            DayGameInfo.Add(gameInfo.Day, TodayGameInfo);

            // 死体で発見されたエージェントが人狼のパターンを削除する
            foreach (Agent agent in gameInfo.LastDeadAgentList)
            {
                ICondition condition = IsMatchPattern.GetIsWolfCondition(agent);
                AllViewSystemInfo.RemovePatterns(condition);
            }

            //TODO 狼全吊りより飽和の方が早い配役の対応。　(15人村→4日目からのため問題なし)(5人村→1狼のため問題無し)
            // 人狼生存人数がおかしいパターンを削除する
            // memo ゲーム終了時は人狼が0人になるかも？
            int maxWolfNum = (gameInfo.AliveAgentList.Count - 1) / 2;
            maxWolfNum = Math.Min(maxWolfNum, GameSetting.RoleNumMap[Role.WEREWOLF]);
            if (gameInfo.Day > GameSetting.RoleNumMap[Role.WEREWOLF])
            {
                IsMatchPattern condition = new IsMatchPattern(warewolves: gameInfo.AliveAgentList);
                AllViewSystemInfo.RemovePatterns(p =>
                {
                    var matchCount = condition.GetMatchCount(p);
                    return matchCount < 1 || matchCount > maxWolfNum;
                });
            }

            // 護衛成功からのパターン削除
            if (gameInfo.GuardedAgent != null && !gameInfo.LastDeadAgentList.Any())
            {
                ICondition condition = IsMatchPattern.GetIsWolfCondition(gameInfo.GuardedAgent);
                AllViewSystemInfo.RemovePatterns(condition);
            }
        }

        public TotalityGameInfo(IGameInfo gameInfo, IGameSetting gameSetting)
        {
            GameSetting = gameSetting;
            latestGameInfo = gameInfo;

            // 全視点情報・自分視点情報の初期化
            InitViewPoint();
        }

        private void InitViewPoint()
        {
            // 全視点システム情報をFactoryから生成
            AllViewSystemInfo = ViewPointFactoryMethod(GameSetting);

            // 全視点CO信用情報を作成
            AllViewTrustInfo = AllViewSystemInfo.Clone();
            AllViewSystemInfo.AddInclusionViewpoint(AllViewTrustInfo);

            // 自分視点システム情報
            SelfViewSystemInfo = AllViewSystemInfo.Clone();
            var SelfWolfSide = new IsMatchPattern(warewolfSide: new List<Agent>() { latestGameInfo.Agent });
            SelfViewSystemInfo.RemovePatterns(p => SelfWolfSide.IsMatch(p));
            AllViewSystemInfo.AddInclusionViewpoint(SelfViewSystemInfo);

            // 自分視点CO信用情報を作成
            SelfViewTrustInfo = SelfViewSystemInfo.Clone();
            SelfViewSystemInfo.AddInclusionViewpoint(SelfViewTrustInfo);
            AllViewTrustInfo.AddInclusionViewpoint(SelfViewTrustInfo);

            // 人外システム情報
            SelfRealRoleSystemtInfo = AllViewSystemInfo.Clone();
            var SelfPossessed = new IsMatchPattern(possessed: new List<Agent>() { latestGameInfo.Agent });
            SelfRealRoleSystemtInfo.RemovePatterns(p => !SelfWolfSide.IsMatch(p));
            AllViewSystemInfo.AddInclusionViewpoint(SelfRealRoleSystemtInfo);

            // 人外CO信用情報
            SelfRealRoleViewTrustInfo = SelfRealRoleSystemtInfo.Clone();
            SelfRealRoleSystemtInfo.AddInclusionViewpoint(SelfRealRoleViewTrustInfo);
            AllViewSystemInfo.AddInclusionViewpoint(SelfRealRoleViewTrustInfo);
        }


        private void UpdateComingOutAndJudge()
        {
            if (!TodayGameInfo.TalkList.Any())
            {
                return;
            }

            foreach (var talk in TodayGameInfo.TalkList.TakeLast(TodayGameInfo.DayTimeAliveAgent.Count()))
            {
                // 直接発話したものの処理
                UpdateComingOutAndJudge(talk, talk.Content);

                // OPERATORを１つ挟んで発言したものの処理
                // TODO OPERATORを２つ以上挟むケースもある？
                if (talk.Content.Operator == Operator.AND || talk.Content.Operator == Operator.DAY)
                {
                    foreach (var content in talk.Content.ContentList)
                    {
                        UpdateComingOutAndJudge(talk, content);
                    }
                }
            }
        }

        private void UpdateComingOutAndJudge(IUtterance talk, Content content)
        {
            // 2019 contest 村騙りによる内訳破壊対策 不自然であろうターンの判定COは無視する
            if (talk.Turn > 4)
            {
                return;
            }

            if (content.Topic == Topic.COMINGOUT && (content.Role == Role.UNC || content.Role == Role.ANY))
            {
                return;
            }
            if (content.Topic == Topic.DIVINED && (content.Result == Species.UNC || content.Result == Species.ANY))
            {
                return;
            }

            if (content.Topic == Topic.COMINGOUT && content.Role.GetTeam() == Team.VILLAGER && content.Target.Equals(talk.Agent))
            {
                AddComingOut(talk.Agent, content.Role, talk);
            }

            if (content.Topic == Topic.DIVINED && (content.Subject.Equals(Agent.NONE) || content.Subject.Equals(talk.Agent)))
            {
                AddComingOut(talk.Agent, Role.SEER, talk);
                AddSeerJudge(talk.Agent, content.Target, content.Result, talk);
            }

            if (content.Topic == Topic.IDENTIFIED && (content.Subject.Equals(Agent.NONE) || content.Subject.Equals(talk.Agent)))
            {
                AddComingOut(talk.Agent, Role.MEDIUM, talk);
                AddSeerJudge(talk.Agent, content.Target, content.Result, talk);
            }
        }

        private void AddComingOut(Agent agent, Role role, IUtterance talk)
        {
            // 同じ役職をＣＯしていれば何もしない
            if (ComingOuts.Any(co => co.IsEnable && co.Agent.Equals(agent) && co.Role.Equals(role)))
            {
                return;
            }

            // 違う役職をＣＯしていれば古いＣＯを無効にする
            bool isCancel = false;
            foreach (var co in ComingOuts.Where(co => co.IsEnable && co.Agent.Equals(agent) && co.Role.Equals(role)))
            {
                co.Cancel(talk);
                isCancel = true;
                SeerJudges.Where(judge => judge.IsEnable && judge.Agent.Equals(agent)).ToList().ForEach(judge => judge.Cancel(talk));
                MediumJudges.Where(judge => judge.IsEnable && judge.Agent.Equals(agent)).ToList().ForEach(judge => judge.Cancel(talk));
            }

            ((List<ComingOut>)ComingOuts).Add(new ComingOut(agent, role, talk));

            if (isCancel)
            {
                // ＣＯ取り消しがあった場合は視点を再構築する
                RebuildViewPoint();
            }
            else
            {
                RemovePatternFromComingOut(role);
            }
        }

        public void RebuildViewPoint()
        {
            AllViewTrustInfo.ResetPatterns(AllViewSystemInfo);
            RemovePatternFromComingOut(Role.SEER);
            RemovePatternFromComingOut(Role.MEDIUM);
        }

        private void RemovePatternFromComingOut(Role role)
        {
            var coAgents = ComingOuts.Where(co => co.IsEnable && co.Role.Equals(role)).ToList();

            // ＣＯの数が本物の数以下（＝騙りが混じっていない可能性がある）なら何もしない
            if (coAgents.Count <= GameSetting.RoleNumMap[role])
            {
                return;
            }

            // 騙りが混じっていないパターンを消す
            IsMatchPattern condition = new IsMatchPattern(warewolfSide: coAgents.Select(co => co.Agent));
            AllViewTrustInfo.RemovePatterns(pattern => condition.GetMatchCount(pattern) <= 0);
        }

        private void AddSeerJudge(Agent agent, Agent target, Species species, IUtterance talk)
        {
            var judge = new Judge(agent, target, species, talk);
            // 判定はそのまま追加（判定騙り→結果取り消しなどは考えない）
            ((List<Judge>)SeerJudges).Add(judge);

            RemovePatternFromJudge(judge);
        }

        private void AddMediumJudge(Agent agent, Agent target, Species species, IUtterance talk)
        {
            var judge = new Judge(agent, target, species, talk);
            // 判定はそのまま追加（判定騙り→結果取り消しなどは考えない）
            ((List<Judge>)MediumJudges).Add(judge);

            RemovePatternFromJudge(judge);
        }

        private void RemovePatternFromJudge(Judge judge)
        {
            // 判定元が真で、判定先が人狼の条件
            IsMatchPattern condition = new IsMatchPattern(notWarewolfSide: new List<Agent> { judge.Agent }, warewolves: new List<Agent>() { judge.Target });
            // 人間判定なら、上記条件を満たす内訳を削除。人狼判定なら、上記条件を満たさない内訳を削除。
            AllViewTrustInfo.RemovePatterns(pattern => condition.IsMatch(pattern) == (judge.Result == Species.HUMAN));
        }

    }
}
