﻿using AIWolf.Lib;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using WhiteWolf.ActionPlan;
using WhiteWolf.Arrangement;
using WhiteWolf.Guess;
using WhiteWolf.Player;
using WhiteWolf.Talk;
using static WhiteWolf.ContestAgent.Statistics;

namespace WhiteWolf.ContestAgent
{
    public class ContestUdonPlayer : AbstractWhiteWolfPlayer
    {

        public override string Name => "Udon";

        private IGuesser guesser;
        protected override IGuesser Guesser => guesser;

        private IActionPlanner actionPlanner;
        protected override IActionPlanner ActionPlanner => actionPlanner;

        ITalkGenerator talkGenerator;
        protected override ITalkGenerator TalkGenerator => talkGenerator;

        ITalkGenerator whisperGenerator = new DummyTalkGenerator();
        protected override ITalkGenerator WhisperGenerator => whisperGenerator;

        private IGameInfo latestGameInfo;

        private LearnData LearnData { get; set; }


        public ContestUdonPlayer()
        {
            // 起動が早いとエージェント番号が若くなって不利になるルールへの対応
            // これでも5人村でほぼ前半になる
            Thread.Sleep(85);
        }

        public override void Initialize(IGameInfo gameInfo, IGameSetting gameSetting)
        {
            try
            {
                // 初回ゲーム開始時に学習オブジェクト初期化
                if (LearnData == null)
                {
                    LearnData = new LearnData(gameInfo.AgentList);
                    guesser = new UdonGuesser(LearnData);
                    actionPlanner = new UdonActionPlanner((UdonGuesser)guesser, LearnData);
                    talkGenerator = new UdonTalkGenerator(LearnData);
                }

                TotallyGameInfo = new TotalityGameInfo(gameInfo, gameSetting);
            }
            catch (Exception ex)
            {
                //Console.WriteLine("Initialize Error.");
            }
        }

        public override void Update(IGameInfo gameInfo)
        {
            try
            {
                latestGameInfo = gameInfo;
                TotallyGameInfo.Update(gameInfo);

                // 村人でも役職を騙るエージェントがCO経由せずに判定出ししていたら、COと判定を取り消す
                var yabaiJudge = TotallyGameInfo.SeerJudges
                    .Concat(TotallyGameInfo.MediumJudges)
                    .Where(judge => TotallyGameInfo.ComingOuts.Any(co => judge.JudgeTalk.Equals(co)) && LearnData.Statistics[judge.Agent][Role.VILLAGER].FakeCO > 0);
                yabaiJudge.ToList().ForEach(judge => judge.Cancel(judge.JudgeTalk));
                TotallyGameInfo.ComingOuts.Where(co => yabaiJudge.Any(judge => co.ComingOutTalk.Equals(judge))).ToList().ForEach(co => co.Cancel(co.ComingOutTalk));
                if (yabaiJudge.Any())
                {
                    TotallyGameInfo.RebuildViewPoint();
                }

                LatestTotalGuess = Guesser.GetGuess(TotallyGameInfo);
                LatestActionPlan = ActionPlanner.GetPlan(TotallyGameInfo, LatestTotalGuess);
            }
            catch (Exception ex)
            {
                //Console.WriteLine("Update Error.");
            }
        }

        public override void Finish()
        {
            try
            {
                var agentList = latestGameInfo.AgentList;
                var winner = latestGameInfo.RoleMap.Any(map => latestGameInfo.AliveAgentList.Contains(map.Key) && map.Value == Role.WEREWOLF) ? Team.WEREWOLF : Team.VILLAGER;
                var roleMap = latestGameInfo.RoleMap;

                roleMap.ToList().ForEach(map => LearnData.Statistics[map.Key][map.Value].GameCount++);
                roleMap.Where(map => map.Value.GetTeam() == winner).ToList().ForEach(map => LearnData.Statistics[map.Key][map.Value].WinCount++);
                roleMap.Where(map => map.Value.GetTeam() != winner).ToList().ForEach(map => LearnData.Statistics[map.Key][map.Value].LoseCount++);

                TotallyGameInfo.ComingOuts.Where(co => co.ComingOutTalk.Day == 1).ToList().ForEach(co =>
                {
                    if (co.Role == Role.SEER) LearnData.Statistics[co.Agent][roleMap[co.Agent]].SeerCO++;
                    if (co.Role == Role.MEDIUM) LearnData.Statistics[co.Agent][roleMap[co.Agent]].MediumCO++;
                    if (roleMap[co.Agent] != co.Role && co.Role != Role.BODYGUARD) LearnData.Statistics[co.Agent][roleMap[co.Agent]].FakeCO++;
                });

                TotallyGameInfo.SeerJudges.Where(judge => judge.JudgeTalk.Day == 1 && judge.Result == Species.HUMAN).ToList()
                    .ForEach(judge => LearnData.Statistics[judge.Agent][roleMap[judge.Agent]].FirstDivineIsWhite++);
                TotallyGameInfo.SeerJudges.Where(judge => judge.JudgeTalk.Day == 1 && judge.Result == Species.WEREWOLF).ToList()
                    .ForEach(judge => LearnData.Statistics[judge.Agent][roleMap[judge.Agent]].FirstDivineIsBlack++);

                // 発話はNOPかANDの先頭のみ見る
                TotallyGameInfo.DayGameInfo.SelectMany(day => day.Value.TalkList).ToList().ForEach(talk =>
                {
                    if (talk.Content.Operator == Operator.NOP) UpdateTalkStatistics(talk.Agent, roleMap[talk.Agent], talk, talk.Content);
                    if (talk.Content.Operator == Operator.AND && talk.Content.ContentList[0].Operator == Operator.NOP) UpdateTalkStatistics(talk.Agent, roleMap[talk.Agent], talk, talk.Content.ContentList[0]);
                });

                foreach (var day in TotallyGameInfo.DayGameInfo.Where(day => day.Value.VoteList.Any()))
                {
                    var voteMap = day.Value.VoteList[0].VoteMap;
                    var ReceiveMap = voteMap.GroupBy(map => map.Value).ToList();
                    foreach (var vote in voteMap)
                    {
                        LearnData.Statistics[vote.Key][roleMap[vote.Key]].Vote++;

                        int cnt = ReceiveMap.SingleOrDefault(g => g.Key.Equals(vote.Value))?.Count() ?? 0;
                        int max = ReceiveMap.Where(g => !g.Key.Equals(vote.Key)).Max(g => g.Count());
                        if (cnt == 1 && voteMap.Any(map => !map.Value.Equals(vote.Key)))
                        {
                            LearnData.Statistics[vote.Key][roleMap[vote.Key]].VoteToNoReceive++;
                        }
                        if (cnt == max)
                        {
                            LearnData.Statistics[vote.Key][roleMap[vote.Key]].VoteToMax++;
                        }
                    }
                }

                UpdatePlayerMeta5();
                UpdatePlayerMeta15();
            }
            catch (Exception ex)
            {
                //Console.WriteLine("Finish Error.");
            }
        }


        private void UpdateTalkStatistics(Agent agent, Role role, IUtterance talk, Content content)
        {
            switch (content.Topic)
            {
                case Topic.Over:
                case Topic.Skip:
                    LearnData.Statistics[agent][role].Talk[talk.Day][talk.Turn][(int)TalkType.OverAndSkip]++;
                    break;
                case Topic.DIVINED:
                case Topic.IDENTIFIED:
                    LearnData.Statistics[agent][role].Talk[talk.Day][talk.Turn][(int)TalkType.Judge]++;
                    break;
                case Topic.COMINGOUT:
                    LearnData.Statistics[agent][role].Talk[talk.Day][talk.Turn][(int)TalkType.ComingOut]++;
                    break;
                case Topic.VOTE:
                    LearnData.Statistics[agent][role].Talk[talk.Day][talk.Turn][(int)TalkType.Vote]++;
                    break;
                case Topic.ESTIMATE:
                    LearnData.Statistics[agent][role].Talk[talk.Day][talk.Turn][(int)TalkType.Estimate]++;
                    break;
                default:
                    LearnData.Statistics[agent][role].Talk[talk.Day][talk.Turn][(int)TalkType.Other]++;
                    break;
            }
        }

        // PL特定(5)
        private void UpdatePlayerMeta5()
        {
            if (TotallyGameInfo.GameSetting.PlayerNum != 5) return;

            var roleMap = latestGameInfo.RoleMap;

            TotallyGameInfo.DayGameInfo[1].TalkList
                .Where(talk => talk.Content.Topic == Topic.DIVINATION)
                .Select(talk => talk.Agent)
                .Distinct()
                .ToList()
                .ForEach(agent =>
                {
//                    Console.WriteLine($"{agent.AgentIdx}\t is \tkenzi.");
                    LearnData.Statistics[agent][roleMap[agent]].Player[(int)PlayerType.kenzi]++;
                });

            foreach (var agent in TotallyGameInfo.AgentList)
            {
                // ここから先は村人10戦以上で判断する
                if (LearnData.Statistics[agent][Role.VILLAGER].GameCount < 10)
                {
                    continue;
                }

                var firstTalkType = new List<int>();
                for (int i = 0; i <= 5; i++)
                {
                    if (LearnData.Statistics[agent][Role.VILLAGER].Talk[1][0][i] > 0)
                    {
                        firstTalkType.Add(i);
                    }
                }

                if (firstTalkType.Count() == 1 && firstTalkType.First() == (int)TalkType.Estimate)
                {
//                    Console.WriteLine($"{agent.AgentIdx}\t is \tyskn67.");
                    LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.yskn67]++;
                    continue;
                }

                if (LearnData.Statistics[agent][Role.VILLAGER].GameCount >= 20 && firstTalkType.Contains((int)TalkType.Vote) && firstTalkType.Contains((int)TalkType.Estimate))
                {
//                    Console.WriteLine($"{agent.AgentIdx}\t is \tfisherman.");
                    LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.fisherman]++;
                    continue;
                }

                if (firstTalkType.Contains((int)TalkType.OverAndSkip) && firstTalkType.Contains((int)TalkType.Estimate))
                {
//                    Console.WriteLine($"{agent.AgentIdx}\t is \twasabi.");
                    LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.wasabi]++;
                    continue;
                }

                if (firstTalkType.Count() == 1 && firstTalkType.First() == (int)TalkType.OverAndSkip)
                {
                    if (LearnData.Statistics[agent][Role.WEREWOLF].FakeCO != LearnData.Statistics[agent][Role.WEREWOLF].GameCount)
                    {
                        if (LearnData.Statistics[agent][Role.WEREWOLF].SeerCO > 0)
                        {
                            if (LearnData.Statistics[agent][Role.VILLAGER].Talk[1][1][(int)TalkType.Vote] > 0)
                            {
//                                Console.WriteLine($"{agent.AgentIdx}\t is \ttakeda.");
                                LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.takeda]++;
                            }
                            else
                            {
//                                Console.WriteLine($"{agent.AgentIdx}\t is \tTomato.");
                                LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.Tomato]++;
                            }
                        }
                        else
                        {
                        }
                    }
                    else
                    {
                        if (LearnData.Statistics[agent][Role.POSSESSED].FirstDivineIsBlack == LearnData.Statistics[agent][Role.POSSESSED].GameCount)
                        {
//                            Console.WriteLine($"{agent.AgentIdx}\t is \tfanfan.");
                            LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.fanfan]++;
                        }
                        else
                        {
//                            Console.WriteLine($"{agent.AgentIdx}\t is \tcalups.");
                            LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.calups]++;

                        }
                    }
                }
                if (firstTalkType.Count() == 1 && firstTalkType.First() == (int)TalkType.Vote)
                {
                    // cantar, GO1DeNO,hello_wolf, Litt1eGirl, sonoda, takaeye

                    // fisherman ここに入ることも
                }
            }
        }

        // PL特定(15)
        private void UpdatePlayerMeta15()
        {
            if (TotallyGameInfo.GameSetting.PlayerNum != 15) return;

            var roleMap = latestGameInfo.RoleMap;

            foreach (var agent in TotallyGameInfo.AgentList)
            {
                // ここから先は10戦以上で判断する
                if (LearnData.Statistics[agent][Role.VILLAGER].GameCount < 10)
                {
                    continue;
                }

                var firstTalkType = new List<int>();
                for (int i = 0; i <= 5; i++)
                {
                    if (LearnData.Statistics[agent][Role.VILLAGER].Talk[1][0][i] > 0)
                    {
                        firstTalkType.Add(i);
                    }
                }

                if (firstTalkType.Count() == 1 && firstTalkType.First() == (int)TalkType.OverAndSkip)
                {
                    switch (firstTalkType.First())
                    {
                        case (int)TalkType.Estimate:
                            if (LearnData.Statistics[agent][Role.SEER].GameCount < 1)
                            {
                                continue;
                            }

                            if (LearnData.Statistics[agent][Role.SEER].Talk[1][2][(int)TalkType.Estimate] > 0)
                            {
//                                Console.WriteLine($"{agent.AgentIdx}\t is \tyskn67.");
                                LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.yskn67]++;
                                continue;
                            }
                            else
                            {
                                // Estimateだけではないが、ほとんどEstimateのためここに到達することもある
//                                Console.WriteLine($"{agent.AgentIdx}\t is \tfisherman.");
                                LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.fisherman]++;
                                continue;
                            }
                    }
                }
            }

            TotallyGameInfo.ComingOuts
                .Where(co => co.Role == Role.BODYGUARD && roleMap[co.Agent] == Role.VILLAGER)
                .ToList()
                .ForEach(co =>
                {
//                    Console.WriteLine($"{co.Agent.AgentIdx}\t is \tLitt1eGirl.");
                    LearnData.Statistics[co.Agent][Role.VILLAGER].Player[(int)PlayerType.Litt1eGirl]++;
                });

            TotallyGameInfo.DayGameInfo.Values
                .SelectMany(day => day.TalkList)
                .Where(talk => talk.Text.Contains("REQUEST"))
                .Select(talk => talk.Agent)
                .Distinct()
                .ToList()
                .ForEach(agent =>
                {
//                    Console.WriteLine($"{agent.AgentIdx}\t is \twasabi.");
                    LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.wasabi]++;
                });
        }

        class DummyTalkGenerator : ITalkGenerator
        {
            public string GetTalkText(ITotalityGameInfo gameInfo, ITotalGuess guess, IActionPlan actionPlan)
            {
                return Content.OVER.Text;
            }
        }

    }
}
