﻿using AIWolf.Lib;
using System;
using System.Collections.Generic;
using System.Linq;
using WhiteWolf.Arrangement;
using WhiteWolf.Guess;
using static WhiteWolf.ContestAgent.Statistics;

namespace WhiteWolf.ContestAgent
{
    public class UdonGuesser : IGuesser
    {
        private LearnData LearnData { get; }

        public UdonGuesser(LearnData learnData)
        {
            LearnData = learnData;
        }

        public ITotalGuess GetGuess(ITotalityGameInfo gameInfo)
        {
            var patterns = GetPatterns(gameInfo);
            var partGuesses = GetPartGuesses(gameInfo);

            // 1人を対象とした推理
            var singleAgentPartGuess = partGuesses.Where(guess => guess is IPartSingleAgentGuess).ToList();
            var singleWolfPatterns = new KeyValuePair<IPattern, double>[gameInfo.GameSetting.PlayerNum + 1];
            var singlePosPatterns = new KeyValuePair<IPattern, double>[gameInfo.GameSetting.PlayerNum + 1];
            foreach (var agent in gameInfo.AgentList)
            {
                singleWolfPatterns[agent.AgentIdx] = new KeyValuePair<IPattern, double>(new Pattern(new List<Agent>() { agent }, new List<Agent>()), 1.0);
                singlePosPatterns[agent.AgentIdx] = new KeyValuePair<IPattern, double>(new Pattern(new List<Agent>(), new List<Agent>() { agent }), 1.0);
            }
            for (int i = 1; i <= gameInfo.GameSetting.PlayerNum; i++)
            {
                foreach (var guess in singleAgentPartGuess)
                {
                    if (guess.IsMatchCondition(singleWolfPatterns[i].Key))
                    {
                        singleWolfPatterns[i] = new KeyValuePair<IPattern, double>(singleWolfPatterns[i].Key, singleWolfPatterns[i].Value * guess.Plausibility);
                    }
                    if (guess.IsMatchCondition(singlePosPatterns[i].Key))
                    {
                        singlePosPatterns[i] = new KeyValuePair<IPattern, double>(singlePosPatterns[i].Key, singlePosPatterns[i].Value * guess.Plausibility);
                    }
                }

            }

            // パターンを対象とした推理
            var multiAgentPartGuess = partGuesses.Where(guess => !(guess is IPartSingleAgentGuess)).ToList();
            var scores = new List<(IPattern, double)>(patterns.Count());
            foreach (var pattern in patterns)
            {
                double score = 1.0;
                pattern.Wolves.ToList().ForEach(a => score *= singleWolfPatterns[a.AgentIdx].Value);
                pattern.Possessed.ToList().ForEach(a => score *= singlePosPatterns[a.AgentIdx].Value);
                // この辺、LINQでIsMatchCondition抽出済みを取ろうとすると妙に重い
                foreach (var guess in multiAgentPartGuess)
                {
                    if (guess.IsMatchCondition(pattern))
                    {
                        score *= guess.Plausibility;
                    }
                }

                scores.Add((pattern, score));
            }
            return new TotalGuess(gameInfo.AgentList, scores);
        }

        public ITotalGuess GetPossessedGuess(ITotalityGameInfo gameInfo)
        {
            var patterns = gameInfo.SelfRealRoleViewTrustInfo.Patterns;
            var partGuesses = GetPossessedPartGuesses(gameInfo);

            if (gameInfo.GameSetting.PlayerNum == 5 || !gameInfo.SelfRealRoleViewTrustInfo.Patterns.Any())
            {
                patterns = gameInfo.SelfRealRoleSystemtInfo.Patterns;
            }

            var scores = new List<(IPattern, double)>(patterns.Count());
            foreach (var pattern in patterns)
            {
                double score = 1.0;
                // この辺、LINQでIsMatchCondition抽出済みを取ろうとすると妙に重い
                foreach (var guess in partGuesses)
                {
                    if (guess.IsMatchCondition(pattern))
                    {
                        score *= guess.Plausibility;
                    }
                }

                scores.Add((pattern, score));
            }
            return new TotalGuess(gameInfo.AgentList, scores);
        }

        /// <summary>
        /// 内訳パターンを取得する。
        /// </summary>
        /// <remarks>
        /// 破綻や村騙りがあるとＣＯを信用した場合の内訳パターンが無いケースがある。
        /// その場合はシステム情報のみから絞り込んだ内訳を使用する。
        /// </remarks>
        private IEnumerable<IPattern> GetPatterns(ITotalityGameInfo gameInfo)
        {
            // 5人村は判定騙りも考慮するためシステム情報からのみ絞り込んだ内訳パターンを使用する
            if (gameInfo.GameSetting.PlayerNum == 5)
            {
                return gameInfo.SelfViewSystemInfo.Patterns;
            }

            // COや判定を信用した視点に内訳があればそれを使用する
            if (gameInfo.SelfViewTrustInfo.Patterns.Any())
            {
                return gameInfo.SelfViewTrustInfo.Patterns;
            }

            // システム情報からのみ絞り込んだ内訳パターンを使用する
            return gameInfo.SelfViewSystemInfo.Patterns;
        }

        List<IPartGuess> firstGuess = new List<IPartGuess>();

        private List<IPartGuess> GetPartGuesses(ITotalityGameInfo gameInfo)
        {
            var partGuesses = new List<IPartGuess>();

            // 初期値はランダムに散らす
            if (gameInfo.TodayGameInfo.Day == 0)
            {
                Random rnd = new Random();
                // 狼狂の値が同じだと内訳がぶれるっぽい
                firstGuess.Clear();
                firstGuess.AddRange(gameInfo.AgentList.Select(agent => new PartSingleAgentWolfSideGuess(agent, 0.9 + rnd.NextDouble() * 0.2)));
            }
            partGuesses.AddRange(firstGuess);

            // 陣形考察
            partGuesses.AddRange(GetFormationGuesses(gameInfo));


            if (gameInfo.GameSetting.PlayerNum == 15)
            {
                // 狼→狼の投票は薄いと見る
                var voteMap = gameInfo.DayGameInfo.Values
                    .Where(dayGameInfo => dayGameInfo.VoteList.Any())
                    .SelectMany(dayGameInfo => dayGameInfo.VoteList[0].VoteMap);
                foreach (var vote in voteMap)
                {
                    var condition = new BothWolfCondition(vote.Key, vote.Value);
                    partGuesses.Add(new ConditionPartGuess(condition, 0.7));
                }

                // 狼→狼の投票は薄いと見る(宣言先)
                var saidVote = new VoteCollection(gameInfo.AliveAgentList, gameInfo.TodayGameInfo.TalkList);
                foreach (var vote in saidVote.VoteMap.Where(vm => vm.Value != null))
                {
                    var condition = new BothWolfCondition(vote.Key, vote.Value);
                    partGuesses.Add(new ConditionPartGuess(condition, 0.7));
                }

                // 狂誤爆、身内切りは薄いとみる
                foreach (var judge in gameInfo.SeerJudges.Where(judge => judge.Result == Species.WEREWOLF && judge.IsEnable))
                {
                    var posCondition = new IsMatchPattern(possessed: new List<Agent>() { judge.Agent }, warewolves: new List<Agent>() { judge.Target });
                    partGuesses.Add(new ConditionPartGuess(posCondition, 0.9));
                    var wolfCondition = new BothWolfCondition(judge.Agent, judge.Target);
                    partGuesses.Add(new ConditionPartGuess(wolfCondition, 0.7));
                }

                // 初手黒は人外多め(多いは多いが、人による)
                partGuesses.AddRange(gameInfo.SeerJudges
                    .Where(judge => judge.JudgeTalk.Day == 1 && judge.Result == Species.WEREWOLF && judge.IsEnable)
                    .Select(judge => new PartSingleAgentWolfSideGuess(judge.Agent, 1.05)));
            }

            // 5人村
            if (gameInfo.GameSetting.PlayerNum == 5)
            {
                // 真占いが村に人狼判定を出す可能性は低い
                foreach (var judge in gameInfo.SeerJudges.Where(judge => judge.Result == Species.WEREWOLF && judge.IsEnable))
                {
                    partGuesses.Add(new FreeConditionPartGuess(pattern => !pattern.Wolves.Concat(pattern.Possessed).Contains(judge.Agent) && !pattern.Wolves.Contains(judge.Target), 0.7));
                }
                // 真占いが人狼に人間判定を出す可能性は低い
                foreach (var judge in gameInfo.SeerJudges.Where(judge => judge.Result == Species.HUMAN && judge.IsEnable))
                {
                    partGuesses.Add(new FreeConditionPartGuess(pattern => !pattern.Wolves.Concat(pattern.Possessed).Contains(judge.Agent) && pattern.Wolves.Contains(judge.Target), 0.3));
                }

                // 自分真占の判定は100%信じる
                foreach (var judge in gameInfo.DayGameInfo.Select(day => day.Value.DivineResult).Where(judge => judge != null))
                {
                    if (judge.Result == Species.WEREWOLF)
                    {
                        partGuesses.Add(new PartSingleAgentNotWolfGuess(judge.Target, 0.0));
                    }
                    else
                    {
                        partGuesses.Add(new PartSingleAgentWolfGuess(judge.Target, 0.0));
                    }
                }

                // 占い候補に投票は人外率高い？
                var seers = gameInfo.ComingOuts.Where(co => co.Role == Role.SEER && co.IsEnable).ToList();
                if (seers.Count < 3 && gameInfo.TodayGameInfo.Day == 1)
                {
                    var saidVote = new VoteCollection(gameInfo.AliveAgentList, gameInfo.TodayGameInfo.TalkList);
                    saidVote.VoteMap.Where(map => seers.Select(co => co.Agent).Contains(map.Value)).Select(map => new PartSingleAgentWolfSideGuess(map.Key, 1.3));
                }

                if (gameInfo.DayGameInfo.ContainsKey(1))
                {
                    var voteCollection = gameInfo.DayGameInfo[1].VoteList.Count > 0 ? gameInfo.DayGameInfo[1].VoteList[0] : new VoteCollection(gameInfo.AgentList, gameInfo.DayGameInfo[1].TalkList);
                    partGuesses.AddRange(voteCollection.VoteMap
                        .Where(map => map.Value != null)
                        .Select(map => new FreeConditionPartGuess(pattern => !pattern.Wolves.Contains(map.Value) && !pattern.Wolves.Concat(pattern.Possessed).Contains(map.Key), 0.8)));
                }
            }

            // 学習
            partGuesses.AddRange(GetLearnGuesses(gameInfo));

            // CCO
            partGuesses.AddRange(gameInfo.ComingOuts
                .Where(co => gameInfo.SeerJudges.Any(judge => co.ComingOutTalk.Day * 100 + co.ComingOutTalk.Turn > judge.JudgeTalk.Day * 100 + judge.JudgeTalk.Turn))
                .Select(co => new PartSingleAgentWolfGuess(co.Agent, 2.0)));

            // 結果数不一致
            foreach (var seerCo in gameInfo.ComingOuts.Where(co => co.IsEnable && co.Role == Role.SEER))
            {
                var coDay = seerCo.ComingOutTalk.Day;
                foreach (var day in gameInfo.DayGameInfo.Where(day => day.Value.TalkList.Count() > day.Value.DayTimeAliveAgent.Count() * 3 && day.Value.DayTimeAliveAgent.Contains(seerCo.Agent)))
                {
                    if (!gameInfo.SeerJudges.Any(judge => judge.Agent.Equals(seerCo.Agent) && judge.JudgeTalk.Day == day.Key))
                    {
                        partGuesses.Add(new PartSingleAgentWolfSideGuess(seerCo.Agent, 3.0));
                    }
                }
            }

            // 噛まれ占いの結果は真で見る？

            return partGuesses;
        }

        private List<IPartGuess> GetLearnGuesses(ITotalityGameInfo gameInfo)
        {
            var partGuesses = new List<IPartGuess>();

            // 学習(占霊CO)
            foreach (var co in gameInfo.ComingOuts.Where(co => co.ComingOutTalk.Day == 1 && co.IsEnable))
            {
                if (co.Role == Role.SEER)
                {
                    var posCount = LearnData.Statistics[co.Agent][Role.POSSESSED].GameCount;
                    var posCO = LearnData.Statistics[co.Agent][Role.POSSESSED].SeerCO;
                    var wolfCount = LearnData.Statistics[co.Agent][Role.WEREWOLF].GameCount;
                    var wolfCO = LearnData.Statistics[co.Agent][Role.WEREWOLF].SeerCO;
                    if (posCount >= 3 && posCO <= 0)
                    {
                        partGuesses.Add(new PartSingleAgentPossessedGuess(co.Agent, 0.6));
                    }
                    if (wolfCount >= 5 && wolfCO <= 0)
                    {
                        partGuesses.Add(new PartSingleAgentWolfGuess(co.Agent, 0.6));
                    }
                }
                if (co.Role == Role.MEDIUM)
                {
                    var posCount = LearnData.Statistics[co.Agent][Role.POSSESSED].GameCount;
                    var posCO = LearnData.Statistics[co.Agent][Role.POSSESSED].MediumCO;
                    var wolfCount = LearnData.Statistics[co.Agent][Role.WEREWOLF].GameCount;
                    var wolfCO = LearnData.Statistics[co.Agent][Role.WEREWOLF].MediumCO;
                    if (posCount >= 3 && posCO <= 0)
                    {
                        partGuesses.Add(new PartSingleAgentPossessedGuess(co.Agent, 0.6));
                    }
                    if (wolfCount >= 5 && wolfCO <= 0)
                    {
                        partGuesses.Add(new PartSingleAgentWolfGuess(co.Agent, 0.6));
                    }
                }
            }
            foreach (var agent in gameInfo.AgentList.Except(gameInfo.ComingOuts.Where(co => co.ComingOutTalk.Day == 1 && co.IsEnable).Select(co => co.Agent)))
            {
                var posCount = LearnData.Statistics[agent][Role.POSSESSED].GameCount;
                var posCO = LearnData.Statistics[agent][Role.POSSESSED].SeerCO + LearnData.Statistics[agent][Role.POSSESSED].MediumCO;
                var wolfCount = LearnData.Statistics[agent][Role.WEREWOLF].GameCount;
                var wolfCO = LearnData.Statistics[agent][Role.WEREWOLF].SeerCO + LearnData.Statistics[agent][Role.WEREWOLF].MediumCO;

                if (posCount >= 3 && posCO >= posCount)
                {
                    partGuesses.Add(new PartSingleAgentPossessedGuess(agent, 0.6));
                }
                if (wolfCount >= 5 && wolfCO >= wolfCount)
                {
                    partGuesses.Add(new PartSingleAgentWolfGuess(agent, 0.6));
                }
            }

            // 学習(初手占いの色)
            foreach (var judge in gameInfo.SeerJudges.Where(judge => judge.JudgeTalk.Day == 1 && judge.IsEnable))
            {
                var seerCount = LearnData.Statistics[judge.Agent][Role.SEER].GameCount;
                var seerBlack = LearnData.Statistics[judge.Agent][Role.SEER].FirstDivineIsBlack;
                var seerWhite = LearnData.Statistics[judge.Agent][Role.SEER].FirstDivineIsWhite;
                var counts = judge.Result == Species.HUMAN ? seerWhite : seerBlack;

                var posCount = LearnData.Statistics[judge.Agent][Role.POSSESSED].GameCount;
                var posBlack = LearnData.Statistics[judge.Agent][Role.POSSESSED].FirstDivineIsBlack;
                var posWhite = LearnData.Statistics[judge.Agent][Role.POSSESSED].FirstDivineIsWhite;
                var countp = judge.Result == Species.HUMAN ? posWhite : posBlack;

                var wolfCount = LearnData.Statistics[judge.Agent][Role.WEREWOLF].GameCount;
                var wolfBlack = LearnData.Statistics[judge.Agent][Role.WEREWOLF].FirstDivineIsBlack;
                var wolfWhite = LearnData.Statistics[judge.Agent][Role.WEREWOLF].FirstDivineIsWhite;
                var countw = judge.Result == Species.HUMAN ? wolfWhite : wolfBlack;

                if (posBlack + posWhite >= 3 && countp == 0)
                {
                    partGuesses.Add(new PartSingleAgentPossessedGuess(judge.Agent, 0.6));
                }
                if (wolfBlack + wolfWhite >= 3 && countw == 0)
                {
                    partGuesses.Add(new PartSingleAgentWolfGuess(judge.Agent, 0.6));
                }

                if (posBlack + posWhite >= 3 && countp == posBlack + posWhite)
                {
                    if (seerBlack + seerWhite >= 3 && countp * 2 <= seerBlack + seerWhite)
                    {
                        partGuesses.Add(new PartSingleAgentPossessedGuess(judge.Agent, 1.5));
                    }
                }
                if (wolfBlack + wolfWhite >= 3 && countp == wolfBlack + wolfWhite)
                {
                    if (seerBlack + seerWhite >= 3 && countp * 2 <= seerBlack + seerWhite)
                    {
                        partGuesses.Add(new PartSingleAgentWolfGuess(judge.Agent, 1.5));
                    }
                }
            }

            // 学習(CO取り消し)
            var cancelCOAgents = gameInfo.ComingOuts.Where(co => !co.IsEnable).Select(co => co.Agent).Distinct();
            foreach (var agent in cancelCOAgents)
            {
                if (LearnData.Statistics[agent][Role.VILLAGER].GameCount > 10 && LearnData.Statistics[agent][Role.VILLAGER].FakeCO <= 0)
                {
                    partGuesses.Add(new PartSingleAgentWolfSideGuess(agent, 1.5));
                }
            }

            // 学習(発言)
            double[] talkWolfRate = Enumerable.Repeat(1.0, 16).ToArray();
            double[] talkPosRate = Enumerable.Repeat(1.0, 16).ToArray();
            foreach (var talk in gameInfo.DayGameInfo.Values.SelectMany(day => day.TalkList))
            {
                Content content = null;
                if (talk.Content.Operator == Operator.NOP) content = talk.Content;
                if (talk.Content.Operator == Operator.AND && talk.Content.ContentList[0].Operator == Operator.NOP) content = talk.Content.ContentList[0];
                if (content == null) continue;

                int type;
                switch (content.Topic)
                {

                    case Topic.Over:
                    case Topic.Skip:
                        type = (int)TalkType.OverAndSkip;
                        break;
                    case Topic.DIVINED:
                    case Topic.IDENTIFIED:
                        type = (int)TalkType.Judge;
                        break;
                    case Topic.COMINGOUT:
                        type = (int)TalkType.ComingOut;
                        break;
                    case Topic.VOTE:
                        type = (int)TalkType.Vote;
                        break;
                    case Topic.ESTIMATE:
                        type = (int)TalkType.Estimate;
                        break;
                    default:
                        type = (int)TalkType.Other;
                        break;
                }

                var agent = talk.Agent;
                Role coRole = Role.VILLAGER;
                if (gameInfo.ComingOuts.Any(co => co.Agent.Equals(agent) && co.IsEnable && co.Role == Role.SEER)) coRole = Role.SEER;
                if (gameInfo.ComingOuts.Any(co => co.Agent.Equals(agent) && co.IsEnable && co.Role == Role.MEDIUM)) coRole = Role.MEDIUM;

                int vilTotalCount = LearnData.Statistics[agent][coRole].Talk[talk.Day][talk.Turn].Sum(typeCount => typeCount);
                int vilTypeCount = LearnData.Statistics[agent][coRole].Talk[talk.Day][talk.Turn][type];

                int wolfTotalCount = LearnData.Statistics[agent][Role.WEREWOLF].Talk[talk.Day][talk.Turn].Sum(typeCount => typeCount);
                int wolfTypeCount = LearnData.Statistics[agent][Role.WEREWOLF].Talk[talk.Day][talk.Turn][type];

                int posTotalCount = LearnData.Statistics[agent][Role.WEREWOLF].Talk[talk.Day][talk.Turn].Sum(typeCount => typeCount);
                int posTypeCount = LearnData.Statistics[agent][Role.WEREWOLF].Talk[talk.Day][talk.Turn][type];

                if (vilTotalCount < 10) continue;
                var vilRate = (vilTypeCount + 50) / (double)(vilTotalCount + 50);

                if (wolfTotalCount >= 5)
                {
                    var wolfRate = (wolfTypeCount + 20) / (double)(wolfTotalCount + 20);
                    talkWolfRate[talk.Agent.AgentIdx] *= Math.Pow(wolfRate / vilRate, 0.5);
                }
                if (posTotalCount >= 5)
                {
                    var posRate = (posTypeCount + 20) / (double)(posTotalCount + 20);
                    talkPosRate[talk.Agent.AgentIdx] *= Math.Pow(posRate / vilRate, 0.5);
                }

            }
            partGuesses.AddRange(talkWolfRate.Select((rate, idx) => new PartSingleAgentWolfGuess(Agent.GetAgent(idx), rate)));
            partGuesses.AddRange(talkPosRate.Select((rate, idx) => new PartSingleAgentPossessedGuess(Agent.GetAgent(idx), rate)));


            // 最多への投票・0票への投票(TODO 当日の宣言分もやりたいけどなぜか重い)
            //var saidVote = new VoteCollection(gameInfo.TodayGameInfo.DayTimeAliveAgent.ToList(), gameInfo.TodayGameInfo.TalkList.ToList());
            var votes = gameInfo.DayGameInfo
                .Where(day => day.Value.VoteList.Any())
                .Select(day => day.Value.VoteList[0])
                //.Append(saidVote)
                .ToList();
            foreach (var voteCollection in votes)
            {
                var voteMap = voteCollection.VoteMap.ToList();
                var ReceiveCountMap = voteMap.GroupBy(map => map.Value).ToDictionary(g => g.Key, g => g.Count());

                foreach (var vote in voteMap)
                {
                    Role coRole = Role.VILLAGER;
                    if (gameInfo.ComingOuts.Any(co => co.Agent.Equals(vote.Key) && co.IsEnable && co.Role == Role.SEER)) coRole = Role.SEER;
                    if (gameInfo.ComingOuts.Any(co => co.Agent.Equals(vote.Key) && co.IsEnable && co.Role == Role.MEDIUM)) coRole = Role.MEDIUM;

                    var vilVote = LearnData.Statistics[vote.Key][coRole].Vote;
                    var vilVoteToMax = LearnData.Statistics[vote.Key][coRole].VoteToMax;
                    var vilVoteToZero = LearnData.Statistics[vote.Key][coRole].VoteToNoReceive;
                    var wolfVote = LearnData.Statistics[vote.Key][Role.WEREWOLF].Vote;
                    var wolfVoteToMax = LearnData.Statistics[vote.Key][Role.WEREWOLF].VoteToMax;
                    var wolfVoteToZero = LearnData.Statistics[vote.Key][Role.WEREWOLF].VoteToNoReceive;
                    var posVote = LearnData.Statistics[vote.Key][Role.POSSESSED].Vote;
                    var posVoteToMax = LearnData.Statistics[vote.Key][Role.POSSESSED].VoteToMax;
                    var posVoteToZero = LearnData.Statistics[vote.Key][Role.POSSESSED].VoteToNoReceive;

                    if (vilVote < 12 || wolfVote < 12 || posVote < 12) continue;

                    int cnt = voteMap.Count(map => map.Value.Equals(vote.Value));
                    int max = ReceiveCountMap.Where(g => !g.Key.Equals(vote.Key)).Max(g => g.Value);
                    var isVoteToMax = (cnt == max);
                    var isVoteToNoReceive = (cnt == 1 && voteMap.Any(map => !map.Value.Equals(vote.Key)));


                    var mVilRate = (vilVoteToMax + 50) / (double)(vilVote + 100);
                    var mWolfRate = (wolfVoteToMax + 25) / (double)(wolfVote + 50);
                    var mPosRate = (posVoteToMax + 25) / (double)(posVote + 50);

                    partGuesses.Add(new PartSingleAgentWolfGuess(vote.Key, isVoteToMax ? (mWolfRate / mVilRate) : (mVilRate / mWolfRate)));
                    partGuesses.Add(new PartSingleAgentPossessedGuess(vote.Key, isVoteToMax ? (mPosRate / mVilRate) : (mVilRate / mPosRate)));

                    var nVilRate = (vilVoteToZero + 50) / (double)(vilVote + 100);
                    var nWolfRate = (wolfVoteToZero + 25) / (double)(wolfVote + 50);
                    var nPosRate = (posVoteToZero + 25) / (double)(posVote + 50);

                    partGuesses.Add(new PartSingleAgentWolfGuess(vote.Key, isVoteToNoReceive ? (nWolfRate / nVilRate) : (nVilRate / nWolfRate)));
                    partGuesses.Add(new PartSingleAgentPossessedGuess(vote.Key, isVoteToNoReceive ? (nPosRate / nVilRate) : (nVilRate / nPosRate)));
                }
            }

            // 固定メタ(事前学習の一種と位置づけ)
            partGuesses.AddRange(GetMetaGuesses5(gameInfo));
            partGuesses.AddRange(GetMetaGuesses15(gameInfo));

            return partGuesses;
        }

        private List<IPartGuess> GetFormationGuesses(ITotalityGameInfo gameInfo)
        {
            var partGuesses = new List<IPartGuess>();

            // 占全偽は薄いと見る
            var seerList = gameInfo.ComingOuts.Where(co => co.IsEnable && co.Role == Role.SEER);
            if (seerList.Any())
            {
                var condition = new IsMatchPattern(warewolfSide: seerList.Select(co => co.Agent));
                partGuesses.Add(new ConditionPartGuess(condition, 0.1));
            }

            // 霊全偽は薄いと見る
            var mediumList = gameInfo.ComingOuts.Where(co => co.IsEnable && co.Role == Role.MEDIUM);
            if (mediumList.Any())
            {
                var condition = new IsMatchPattern(warewolfSide: mediumList.Select(co => co.Agent));
                partGuesses.Add(new ConditionPartGuess(condition, 0.1));
            }

            // 狂人は占COしていると見る
            if (seerList.Count() >= 2)
            {
                var condition = new IsMatchPattern(possessed: seerList.Select(co => co.Agent));
                partGuesses.Add(new FreeConditionPartGuess(pattern => condition.GetMatchCount(pattern) <= 0, 0.5));
            }

            // 5人村占3COなら人狼が占COしていると見る(村騙りエージェントがいる場合は基準人数を増やす)
            var wolfBoys = LearnData.Statistics.Where(agent => agent.Value[Role.VILLAGER].FakeCO > 0).Select(agent => agent.Key).ToList();
            if (gameInfo.GameSetting.PlayerNum == 5 && seerList.Count() >= 3 + wolfBoys.Count())
            {
                var condition = new IsMatchPattern(warewolves: seerList.Select(co => co.Agent));
                partGuesses.Add(new FreeConditionPartGuess(pattern => condition.GetMatchCount(pattern) <= 0, 0.1));
            }

            return partGuesses;
        }

        /// <summary>
        /// 狂人が狂人視点で使う推理
        /// </summary>
        public List<IPartGuess> GetPossessedPartGuesses(ITotalityGameInfo gameInfo)
        {
            var partGuesses = new List<IPartGuess>();

            partGuesses.AddRange(GetFormationGuesses(gameInfo));
            partGuesses.AddRange(GetLearnGuesses(gameInfo));

            var seerList = gameInfo.ComingOuts.Where(co => co.IsEnable && co.Role == Role.SEER && co.IsEnable);
            var mediumList = gameInfo.ComingOuts.Where(co => co.IsEnable && co.Role == Role.MEDIUM && co.IsEnable);

            // CCO
            partGuesses.AddRange(gameInfo.ComingOuts
                .Where(co => gameInfo.SeerJudges.Any(judge => co.ComingOutTalk.Day * 100 + co.ComingOutTalk.Turn > judge.JudgeTalk.Day * 100 + judge.JudgeTalk.Turn))
                .Select(co => new FreeConditionPartGuess(pattern => !pattern.Wolves.Contains(co.Agent), 2.0)));

            // 5人村
            if (gameInfo.GameSetting.PlayerNum == 5)
            {
                // 真占いが村に人狼判定を出す可能性は低い
                foreach (var judge in gameInfo.SeerJudges.Where(judge => judge.Result == Species.WEREWOLF && judge.IsEnable))
                {
                    partGuesses.Add(new FreeConditionPartGuess(pattern => !pattern.Wolves.Concat(pattern.Possessed).Contains(judge.Agent) && !pattern.Wolves.Contains(judge.Target), 0.7));
                }
                // 真占いが人狼に人間判定を出す可能性は低い
                foreach (var judge in gameInfo.SeerJudges.Where(judge => judge.Result == Species.HUMAN && judge.IsEnable))
                {
                    partGuesses.Add(new FreeConditionPartGuess(pattern => !pattern.Wolves.Concat(pattern.Possessed).Contains(judge.Agent) && pattern.Wolves.Contains(judge.Target), 0.3));
                }
                // 5人村占3COなら人狼が占COしていると見る(村騙りエージェントがいる場合は基準人数を増やす)
                var wolfBoys = LearnData.Statistics.Where(agent => agent.Value[Role.VILLAGER].FakeCO > 0).Select(agent => agent.Key).ToList();
                if (gameInfo.GameSetting.PlayerNum == 5 && seerList.Count() >= 3 + wolfBoys.Count())
                {
                    var condition = new IsMatchPattern(warewolves: seerList.Select(co => co.Agent));
                    partGuesses.Add(new FreeConditionPartGuess(pattern => condition.GetMatchCount(pattern) <= 0, 0.5));
                }
                // 占い候補に投票は人外率高い？
                var seers = gameInfo.ComingOuts.Where(co => co.Role == Role.SEER && co.IsEnable).ToList();
                if (seers.Count < 3 && gameInfo.TodayGameInfo.Day == 1)
                {
                    var saidVote = new VoteCollection(gameInfo.AliveAgentList, gameInfo.TodayGameInfo.TalkList);
                    saidVote.VoteMap.Where(map => seers.Select(co => co.Agent).Contains(map.Value)).Select(map => new PartSingleAgentWolfSideGuess(map.Key, 1.3));
                }
            }
            else
            {
                // 狼→狼の投票は薄いと見る
                var voteMap = gameInfo.DayGameInfo.Values
                    .Where(dayGameInfo => dayGameInfo.VoteList.Any())
                    .SelectMany(dayGameInfo => dayGameInfo.VoteList[0].VoteMap);
                foreach (var vote in voteMap)
                {
                    var condition = new BothWolfCondition(vote.Key, vote.Value);
                    partGuesses.Add(new ConditionPartGuess(condition, 0.7));
                }
            }

            return partGuesses;
        }

        private List<IPartGuess> GetMetaGuesses5(ITotalityGameInfo gameInfo)
        {
            var partGuesses = new List<IPartGuess>();

            if (gameInfo.GameSetting.PlayerNum != 5)
            {
                return partGuesses;
            }

            if (gameInfo.TodayGameInfo.Day < 1)
            {
                return partGuesses;
            }

            // takeda seerCOでd1t3がskipなら偽
            var takeda = gameInfo.ComingOuts
                .Where(co => co.Role == Role.SEER && co.IsEnable && gameInfo.DayGameInfo[1].TalkList.Any(talk => talk.Agent.Equals(co.Agent) && talk.Content.Topic == Topic.Skip && talk.Turn == 3))
                .Where(co => LearnData.Statistics[co.Agent][Role.VILLAGER].Player[(int)PlayerType.takeda] > 0)
                .ToList();
            partGuesses.AddRange(takeda.Select(co => new PartSingleAgentWolfSideGuess(co.Agent, 2.0)));

            // takeda 初手白判定ならpossessed
            var takeda2 = gameInfo.SeerJudges
                .Where(judge => judge.IsEnable && judge.JudgeTalk.Day == 1 && judge.Result == Species.HUMAN)
                .Where(judge => LearnData.Statistics[judge.Agent][Role.VILLAGER].Player[(int)PlayerType.takeda] > 0)
                .ToList();
            partGuesses.AddRange(takeda.Select(co => new PartSingleAgentPossessedGuess(co.Agent, 1.8)));

            // fanfan seer coでestimateなしが狂、白判定が狼
            var fanfan = gameInfo.AgentList.Where(agent => LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.fanfan] > 0).ToList();
            foreach (var agent in fanfan)
            {
                if (!gameInfo.ComingOuts.Any(co => co.Agent.Equals(agent))) { partGuesses.Add(new PartSingleAgentWolfSideGuess(agent, 0.8)); }
                else if (gameInfo.SeerJudges.Any(judge => judge.Agent.Equals(agent) && judge.Result == Species.HUMAN)) { partGuesses.Add(new PartSingleAgentWolfGuess(agent, 1.8)); }
                else if (gameInfo.DayGameInfo[1].TalkList.Any(talk => talk.Agent.Equals(agent) && talk.Content.Topic == Topic.ESTIMATE)) { partGuesses.Add(new PartSingleAgentWolfSideGuess(agent, 0.8)); }
                else { partGuesses.Add(new PartSingleAgentPossessedGuess(agent, 1.5)); }
            }

            // Tomato seerco 1t skip は偽
            var tomato = gameInfo.ComingOuts
                .Where(co => co.IsEnable && co.Role == Role.SEER)
                .Where(co => LearnData.Statistics[co.Agent][Role.VILLAGER].Player[(int)PlayerType.Tomato] > 0)
                .Where(co => gameInfo.DayGameInfo[1].TalkList.Any(talk => talk.Agent.Equals(co.Agent) && talk.Turn == 1 && talk.Content.Topic == Topic.Skip))
                .ToList();
            partGuesses.AddRange(tomato.Select(co => new PartSingleAgentPossessedGuess(co.Agent, 1.4)));

            // 0804もういない？
            // 素村で毎回黒出しする人の村COは人狼
            partGuesses.AddRange(gameInfo.ComingOuts
                .Where(co => co.Role == Role.VILLAGER && LearnData.Statistics[co.Agent][Role.VILLAGER].GameCount > 3 && LearnData.Statistics[co.Agent][Role.VILLAGER].GameCount == LearnData.Statistics[co.Agent][Role.VILLAGER].FirstDivineIsBlack)
                .Select(co => new PartSingleAgentWolfGuess(co.Agent, 1.2)));

            // 占COで村陣営予想は非狼が多い
            partGuesses.AddRange(gameInfo.DayGameInfo[1].TalkList
                .Where(talk => talk.Content.Topic == Topic.ESTIMATE && (talk.Content.Role == Role.VILLAGER || talk.Content.Role == Role.SEER))
                .Select(talk => talk.Agent)
                .Distinct()
                .Where(agent => gameInfo.ComingOuts.Any(co => co.Agent.Equals(agent) && co.Role == Role.SEER && co.IsEnable))
                .Select(agent => new PartSingleAgentWolfGuess(agent, 0.9)));

            // 占COで狼陣営予想は非狼が多い
            partGuesses.AddRange(gameInfo.DayGameInfo[1].TalkList
                .Where(talk => talk.Content.Topic == Topic.ESTIMATE && (talk.Content.Role != Role.VILLAGER && talk.Content.Role != Role.SEER))
                .Select(talk => talk.Agent)
                .Distinct()
                .Where(agent => gameInfo.ComingOuts.Any(co => co.Agent.Equals(agent) && co.Role == Role.SEER && co.IsEnable))
                .Select(agent => new PartSingleAgentWolfGuess(agent, 0.8)));

            // 村COで同意は非狼が多い
            partGuesses.AddRange(gameInfo.DayGameInfo[1].TalkList
                .Where(talk => talk.Content.Topic == Topic.AGREE)
                .Select(talk => talk.Agent)
                .Distinct()
                .Where(agent => gameInfo.ComingOuts.Any(co => co.Agent.Equals(agent) && co.Role == Role.VILLAGER && co.IsEnable))
                .Select(agent => new PartSingleAgentWolfGuess(agent, 0.95)));

            // 初手狼COは人外
            partGuesses.AddRange(gameInfo.DayGameInfo[1].TalkList
                .Where(talk => talk.Turn == 0 && talk.Content.Topic == Topic.COMINGOUT && talk.Content.Role == Role.WEREWOLF)
                .Select(talk => talk.Agent)
                .Distinct()
                .Select(agent => new PartSingleAgentWolfSideGuess(agent, 1.8)));

            return partGuesses;
        }


        private List<IPartGuess> GetMetaGuesses15(ITotalityGameInfo gameInfo)
        {
            var partGuesses = new List<IPartGuess>();

            if (gameInfo.GameSetting.PlayerNum != 15)
            {
                return partGuesses;
            }

            if (gameInfo.TodayGameInfo.Day < 1)
            {
                return partGuesses;
            }

            // yskn67 fisherman 狼の狼予想はほぼ村人相手
            var ysknfisher = gameInfo.AgentList
                .Where(agent => LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.yskn67] > 0 || LearnData.Statistics[agent][Role.VILLAGER].Player[(int)PlayerType.fisherman] > 0)
                .ToList();
            partGuesses.AddRange(gameInfo.DayGameInfo.Values
                .SelectMany(day => day.TalkList)
                .Where(talk => ysknfisher.Contains(talk.Agent) && talk.Content.Topic == Topic.ESTIMATE && talk.Content.Role == Role.WEREWOLF)
                .Select(talk => new FreeConditionPartGuess(pattern => pattern.Wolves.Contains(talk.Agent) && !pattern.Wolves.Contains(talk.Content.Target), 0.6)));

            //  L1ttleGirl 狩CO多すぎは狼
            foreach (var agent in gameInfo.AgentList)
            {
                var co = gameInfo.DayGameInfo
                    .SelectMany(day => day.Value.TalkList)
                    .Count(talk => talk.Agent.Equals(agent) && talk.Content.Topic == Topic.COMINGOUT && talk.Content.Role == Role.BODYGUARD);
                if (co > 4)
                {
                    partGuesses.Add(new PartSingleAgentWolfGuess(agent, 2.0));
                }
            }

            // calups seercoで1d0tにskipは狼
            var seerList = gameInfo.ComingOuts.Where(co => co.IsEnable && co.Role == Role.SEER && co.IsEnable);
            partGuesses.AddRange(gameInfo.DayGameInfo[1].TalkList
                .Where(talk => talk.Turn == 0 && talk.Content.Topic == Topic.Skip && seerList.Any(co => co.Agent.Equals(talk.Agent)))
                .Select(talk => new PartSingleAgentWolfGuess(talk.Agent, 1.8)));

            // cantar seercoで1d0tにvoteは狼(他エージェント巻き込み注意)

            // fanfan unc で1d1tにskipは狼が多い

            return partGuesses;
        }



    }
}