﻿using AIWolf.Lib;
using System;
using System.Collections.Generic;
using System.Linq;
using WhiteWolf.Arrangement;

namespace WhiteWolf.Guess
{
    public class SampleGuesser : IGuesser
    {
        public ITotalGuess GetGuess(ITotalityGameInfo gameInfo)
        {
            var patterns = GetPatterns(gameInfo);
            var partGuesses = GetPartGuesses(gameInfo);

            var scores = new List<(IPattern, double)>(patterns.Count());
            foreach (var pattern in patterns)
            {
                double score = 1.0;
                // この辺、LINQでIsMatchCondition抽出済みを取ろうとすると妙に重い
                foreach (var guess in partGuesses)
                {
                    if (guess.IsMatchCondition(pattern))
                    {
                        score *= guess.Plausibility;
                    }
                }

                scores.Add((pattern, score));
            }
            return new TotalGuess(gameInfo.AgentList, scores);
        }

        /// <summary>
        /// 内訳パターンを取得する。
        /// </summary>
        /// <remarks>
        /// 破綻や村騙りがあるとＣＯを信用した場合の内訳パターンが無いケースがある。
        /// その場合はシステム情報のみから絞り込んだ内訳を使用する。
        /// </remarks>
        private IEnumerable<IPattern> GetPatterns(ITotalityGameInfo gameInfo)
        {
            // 5人村は判定騙りも考慮するためシステム情報からのみ絞り込んだ内訳パターンを使用する
            if (gameInfo.GameSetting.PlayerNum == 5)
            {
                return gameInfo.SelfViewSystemInfo.Patterns;
            }

            // COや判定を信用した視点に内訳があればそれを使用する
            if (gameInfo.SelfViewTrustInfo.Patterns.Any())
            {
                return gameInfo.SelfViewTrustInfo.Patterns;
            }

            // システム情報からのみ絞り込んだ内訳パターンを使用する
            return gameInfo.SelfViewSystemInfo.Patterns;
        }

        List<IPartGuess> firstGuess = new List<IPartGuess>();

        private List<IPartGuess> GetPartGuesses(ITotalityGameInfo gameInfo)
        {
            var partGuesses = new List<IPartGuess>();

            // 初期値はランダムに散らす
            if (!firstGuess.Any())
            {
                Random rnd = new Random();
                // 狼狂の値が同じだと内訳がぶれるっぽい
                firstGuess.AddRange(gameInfo.AgentList.Select(agent => new ConditionPartGuess(new IsMatchPattern(warewolfSide: new List<Agent>() { agent }), 0.9 + rnd.NextDouble() * 0.2)));
            }
            partGuesses.AddRange(firstGuess);

            // 占全偽は薄いと見る
            var seerList = gameInfo.ComingOuts.Where(co => co.IsEnable && co.Role == Role.SEER);
            if (seerList.Any())
            {
                var condition = new IsMatchPattern(warewolfSide: seerList.Select(co => co.Agent));
                partGuesses.Add(new ConditionPartGuess(condition, 0.1));
            }

            // 霊全偽は薄いと見る
            var mediumList = gameInfo.ComingOuts.Where(co => co.IsEnable && co.Role == Role.MEDIUM);
            if (mediumList.Any())
            {
                var condition = new IsMatchPattern(warewolfSide: mediumList.Select(co => co.Agent));
                partGuesses.Add(new ConditionPartGuess(condition, 0.1));
            }

            // 狂人は占か霊COしていると見る
            if (seerList.Any() || mediumList.Any())
            {
                var condition = new IsMatchPattern(possessed: mediumList.Concat(seerList).Select(co => co.Agent));
                partGuesses.Add(new FreeConditionPartGuess(pattern => condition.GetMatchCount(pattern) <= 0, 0.5));
            }

            // 狼→狼の投票は薄いと見る
            var voteMap = gameInfo.DayGameInfo.Values
                .Where(dayGameInfo => dayGameInfo.VoteList.Any())
                .SelectMany(dayGameInfo => dayGameInfo.VoteList[0].VoteMap);
            foreach (var vote in voteMap)
            {
                var condition = new BothWolfCondition(vote.Key, vote.Value);
                partGuesses.Add(new ConditionPartGuess(condition, 0.7));
            }

            // 狼→狼の投票は薄いと見る(宣言先)
            var saidVote = new VoteCollection(gameInfo.AliveAgentList, gameInfo.TodayGameInfo.TalkList);
            foreach (var vote in saidVote.VoteMap.Where(vm => vm.Value != null))
            {
                var condition = new BothWolfCondition(vote.Key, vote.Value);
                partGuesses.Add(new ConditionPartGuess(condition, 0.7));
            }

            // 狂誤爆、身内切りは薄いとみる
            if (gameInfo.GameSetting.PlayerNum == 15)
            {
                foreach (var judge in gameInfo.SeerJudges.Where(judge => judge.Result == Species.WEREWOLF))
                {
                    var posCondition = new IsMatchPattern(possessed: new List<Agent>() { judge.Agent }, warewolves: new List<Agent>() { judge.Target });
                    partGuesses.Add(new ConditionPartGuess(posCondition, 0.9));
                    var wolfCondition = new BothWolfCondition(judge.Agent, judge.Target);
                    partGuesses.Add(new ConditionPartGuess(wolfCondition, 0.7));
                }
            }

            // 5人村
            if (gameInfo.GameSetting.PlayerNum == 5)
            {
                // 真占いが村に人狼判定を出す可能性は低い
                foreach (var judge in gameInfo.SeerJudges.Where(judge => judge.Result == Species.WEREWOLF))
                {
                    partGuesses.Add(new FreeConditionPartGuess(pattern => !pattern.Wolves.Concat(pattern.Possessed).Contains(judge.Target) && !pattern.Wolves.Contains(judge.Target), 0.4));
                }
                // 真占いが人狼に人間判定を出す可能性は低い
                foreach (var judge in gameInfo.SeerJudges.Where(judge => judge.Result == Species.HUMAN))
                {
                    partGuesses.Add(new FreeConditionPartGuess(pattern => !pattern.Wolves.Concat(pattern.Possessed).Contains(judge.Target) && pattern.Wolves.Contains(judge.Target), 0.4));
                }

                // 自分真占の判定は100%信じる
                foreach (var judge in gameInfo.DayGameInfo.Select(day => day.Value.DivineResult).Where(judge => judge != null))
                {
                    if (judge.Result == Species.WEREWOLF)
                    {
                        partGuesses.Add(new FreeConditionPartGuess(pattern => !pattern.Wolves.Contains(judge.Target), 0.0));
                    }
                    else
                    {
                        partGuesses.Add(new FreeConditionPartGuess(pattern => pattern.Wolves.Contains(judge.Target), 0.0));

                    }
                }
            }


            // 噛まれ占いの結果は真で見る
            // 初手黒は狂人？


            return partGuesses;
        }

    }
}
